"use strict";

exports.__esModule = true;
exports.createHooks = createHooks;
exports.useSelector = exports.useDispatch = exports.useAtom = exports.useActions = void 0;
var _react = require("react");
var _context = require("./context");
var _differs = require("./differs");
var _raf = require("./raf");
var isServer = typeof navigator === 'undefined';
var identity = function identity(x) {
  return x;
};
var immediate = function immediate(fn) {
  return fn();
};
var delayed = function delayed(fn) {
  return (0, _raf.raf)(fn)();
};
var i = 0;
var nextOrder = function nextOrder() {
  return ++i;
};
function createHooks(AtomContext) {
  function useSelector(selectorFn, options) {
    if (selectorFn === void 0) {
      selectorFn = identity;
    }
    if (options === void 0) {
      options = {};
    }
    var _options = options,
      _options$sync = _options.sync,
      sync = _options$sync === void 0 ? false : _options$sync,
      _options$observe = _options.observe,
      observe = _options$observe === void 0 ? !isServer : _options$observe;
    var atom = useAtom();
    assert(atom, 'No atom found in context, did you forget to wrap your app in <Provider atom={atom} />?');

    // cache the schedule and selector functions
    var schedule = (0, _react.useCallback)(sync ? immediate : delayed, [sync]);
    var selector = (0, _react.useCallback)(selectorFn, options.deps || []);

    // we use a state to trigger a rerender when relevant atom
    // state changes, we don't atom the actual mapped atom state
    // here, because that is only 1 of 2 ways that the component
    // gets rerendered, the other way is being rerended by parent
    var _useState = (0, _react.useState)({}),
      rerender = _useState[1];

    // keep track of rendering order, this is important for:
    // - correctness – parent must rerender first
    // - performance – parent rerendering children should cancel children's scheduled rerenders
    var order = (0, _react.useRef)();

    // keep last used props here for diffing upon each change
    var mappedProps = (0, _react.useRef)();

    // for cancelling scheduled updates in case of parent renders
    var cancelUpdate = (0, _react.useRef)(null);
    if (!order.current) {
      order.current = nextOrder();
    }

    // atom current mapped state on each render
    // so we can diff when atom triggers callbacks
    mappedProps.current = selector(atom.get());

    // cancel any pending scheduled updates after each render
    // since we just go rerendered by the parent component
    invoke(cancelUpdate);
    (0, _react.useEffect)(function observe() {
      if (!observe) return;

      // very important to check for this, since
      // our observe callback might have been removed
      // from the atom's listeners array while atom is
      // looping over the old list of listener references
      var didUnobserve = false;
      var unobserve = atom.observe(onChange, order.current);

      // avoid race render/commit phase conditions
      // trigger this to check if atom's state change before
      // we managed to subscribe in this effect
      onChange();
      function onChange() {
        if (didUnobserve) return;

        // take into account atom updates happening in rapid sequence
        // cancel each previously scheduled one and reschedule
        invoke(cancelUpdate);

        // schedule an update
        cancelUpdate.current = schedule(function scheduledOnChange() {
          cancelUpdate.current = null;
          var nextMappedProps = selector(atom.get());
          if ((0, _differs.differs)(mappedProps.current, nextMappedProps)) {
            rerender({});
          }
        });
      }
      return function destroy() {
        didUnobserve = true;
        unobserve();
        invoke(cancelUpdate);
      };
    }, [atom, observe, selector, schedule, order, mappedProps, cancelUpdate, rerender]);

    // always return fresh mapped props, in case
    // this is a parent rerendering children
    return mappedProps.current;
  }
  function useActions() {
    var atom = useAtom();
    return atom && atom.actions;
  }
  function useDispatch() {
    var atom = useAtom();
    return atom && atom.dispatch;
  }
  function useAtom() {
    var _useContext = (0, _react.useContext)(AtomContext),
      atom = _useContext.atom;
    return atom;
  }
  return {
    useSelector: useSelector,
    useActions: useActions,
    useDispatch: useDispatch,
    useAtom: useAtom
  };
}
function assert(cond, error) {
  if (!cond) {
    throw new Error(error);
  }
}
function invoke(ref) {
  if (ref.current) {
    ref.current();
    ref.current = null;
  }
}
var hooks = createHooks(_context.AtomContext);
var useAtom = exports.useAtom = hooks.useAtom;
var useSelector = exports.useSelector = hooks.useSelector;
var useActions = exports.useActions = hooks.useActions;
var useDispatch = exports.useDispatch = hooks.useDispatch;